/* -*- C++ -*-
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Interface to the WIND target abstraction.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _WindNode_h
#define _WindNode_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "CkNode.h"
#include "vxworks/vxWorks.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_WIND
#include "sodecl.h"

#define WIND_MAGIC  0x57494e44	// WIND kernel magic number

#define WIND_TASK_INODE   0
#define WIND_MQ_INODE     1
#define WIND_SEMA4_INODE  2
#define WIND_MUTEX_INODE  3
#define WIND_WD_INODE     4

#define WindRunningNode   ((WindNode *)CkNode::This)
#define WindRunningTask   ((WindTask *)WindRunningNode->getRunningThread())

#define WIND_SIG_INTR     0
#define WIND_SIG_RESTART  1

typedef int wind_tid_t;

struct WindEventHook : public Link {

    FUNCPTR fn;

    WindEventHook(FUNCPTR _fn) {
	fn = _fn;
    }
};

MakeList(WindEventHook);

class WindTask;
class WindMutex;
class WindSema4;
class WindMsgQ;
class WindWDog;

class WindNode : public CkNode {

    friend class WindConfig;

protected:

    int schedLocks;

    int nslice;

    void (*startAddr)(void);

    unsigned customOptions;

    CkInterrupt *rtc;
    
    WindEventHookList createHookList,
	switchHookList,
	deleteHookList;

    WindTask *defaultTask;

    WindTask *findTaskByTcb(WIND_TCB *pTcb);

    WindSema4 *findSemByHandle(CkHandle handle);

    WindMutex *findMutexByHandle(CkHandle handle);

    WindMsgQ *findQueueByHandle(CkHandle handle);

    WindWDog *findWDogByHandle(CkHandle handle);

    int dispatchSignals(WindTask *task);

public:

    WindNode(const char *nodeName,
	     CkNodeParms *nparms);

    int testOption(int option) const {
	return !!(customOptions & option);
    }

    void setContextErrno(int errorNumber);

    void nameObject(CkObject *object,
		    int type);

    void nameObject(CkObject *object,
		    int type,
		    CkExtId uname);

    void unnameObject(CkObject *object);

    CkObject *findActiveObject(CkId id);

    int inTaskContext() const {
	return !isAsynch() &&
	    SxThread::currentThread->protoTestStatus(CK_RUNNABLE_MARKER) != 0;
    }

    void rtcEnable() {
	rtc->clrStatus(CK_INTR_MASKED);
    } 

    void rtcDisable() {
	rtc->setStatus(CK_INTR_MASKED);
    }

    void rtcProgram(double ticksPerSec) {
	rtc->configure(CfEventPeriodical,
		       CString().format("%f msc",1000.0/ticksPerSec));
    }

    STATUS checkTaskContext (const char *syscall);

    STATUS kdoor(kernelTimeSlice)(int ticks);

    void kdoor(tickSet)(ckticks_t ticks);

    ckticks_t kdoor(tickGet)(void);

    STATUS kdoor(taskActivate)(WIND_TCB *pTcb);

    STATUS kdoor(taskCreateHookAdd)(FUNCPTR createHook);

    STATUS kdoor(taskSwitchHookAdd)(FUNCPTR switchHook);

    STATUS kdoor(taskDeleteHookAdd)(FUNCPTR deleteHook);

    STATUS kdoor(taskCreateHookDelete)(FUNCPTR createHook);

    STATUS kdoor(taskSwitchHookDelete)(FUNCPTR switchHook);

    STATUS kdoor(taskDeleteHookDelete)(FUNCPTR deleteHook);

    STATUS kdoor(taskDelay)(int ticks);

    STATUS kdoor(taskDelete)(WIND_TCB *pTcb);

    STATUS kdoor(taskDeleteForce)(WIND_TCB *pTcb);

    void kdoor(taskExit)(int status);

    WIND_TCB *kdoor(taskIdSelf)(void);

    STATUS kdoor(taskIdVerify)(WIND_TCB *pTcb);

    STATUS kdoor(taskInit)(WIND_TCB *pTcb,
			   const char *suffix,
			   int priority,
			   int options,
			   int stackSize,
			   FUNCPTR entryPt,
			   int args[10],
			   WIND_TCB **ppTcb);

    STATUS kdoor(taskLock)(void);

    STATUS kdoor(taskPrioritySet)(WIND_TCB *pTcb,
				  int priority);

    STATUS kdoor(taskPriorityGet)(WIND_TCB *pTcb,
				  int *pPriority);

    STATUS kdoor(taskRestart)(WIND_TCB *pTcb);

    STATUS kdoor(taskResume)(WIND_TCB *pTcb);

    STATUS kdoor(taskSafe)(void);

    WIND_TCB *kdoor(taskSpawn)(const char *suffix,
			       int priority,
			       int options,
			       int stackSize,
			       FUNCPTR entryPt,
			       int args[10]);

    STATUS kdoor(taskSuspend)(WIND_TCB *pTcb);

    STATUS kdoor(taskUnlock)(void);

    STATUS kdoor(taskUnsafe)(void);

    WIND_TCB *kdoor(taskTcb)(WIND_TCB *pTcb);

    CkHandle kdoor(semBCreate)(int options,
			       SEM_B_STATE initialState);

    CkHandle kdoor(semCCreate)(int options,
			       int initialCount);

    CkHandle kdoor(semMCreate)(int options);

    STATUS kdoor(semDelete)(CkHandle handle);

    STATUS kdoor(semFlush)(CkHandle handle);

    STATUS kdoor(semTake)(CkHandle handle,
			  int ticks);

    STATUS kdoor(semGive)(CkHandle handle);

    CkHandle kdoor(msgQCreate)(int maxMsg,
			       unsigned maxLength,
			       int options);

    STATUS kdoor(msgQDelete)(CkHandle handle);

    STATUS kdoor(msgQSend)(CkHandle handle,
			   char *buffer,
			   unsigned nBytes,
			   int priority,
			   int ticks);

    int kdoor(msgQReceive)(CkHandle handle,
			   char *buffer,
			   unsigned nBytes,
			   int ticks);

    int kdoor(msgQNumMsgs)(CkHandle handle);

    CkHandle kdoor(wdCreate)(void);

    STATUS kdoor(wdDelete)(CkHandle handle);

    STATUS kdoor(wdStart)(CkHandle handle,
			  int ticks,
			  FUNCPTR pRoutine,
			  int param);

    STATUS kdoor(wdCancel)(CkHandle handle);

    WIND_TCB *kdoor(taskIdDefault)(WIND_TCB *pTcb);

    const char *kdoor(taskName)(WIND_TCB *pTcb);

    WIND_TCB *kdoor(taskNameToId)(const char *name);

    int kdoor(taskIsReady)(WIND_TCB *pTcb);

    int kdoor(taskIsSuspended)(WIND_TCB *pTcb);

    virtual ckticks_t rrGetCredit(void);

    virtual void (*getRootEntry())(...);

    virtual void protoInit(void);

    virtual void kinit(warmInit)(void);

    void kcout(fireCallout)(CkEventKind event,
			    CkObject *object);

    virtual CkInterrupt *createInterrupt (ckvector_t vector,
					  cklevel_t level,
					  ckisr_t *isr,
					  ckdsr_t *dsr,
					  void *cookie,
					  const char *symbolicName);
};

MakeGList(WindNode);

class WindWDog : public CkAlarm {

protected:

    FUNCPTR fn;
    int arg;

    WindNode *getNode() {
	return (WindNode *)CkAlarm::getNode();
    }

public:

    WindWDog(WindNode *node);

    virtual ~WindWDog();

    void start(FUNCPTR fn,
	       int arg,
	       ckticks_t ticks);

    void stop(void);

    void kroot(trampoline)() {
	fn(arg);
    }

    virtual void timeout(CkObject *client) {
	// The indirection through the trampoline is aimed at
	// interleaving a kroot() frame so the debugger will
	// consider the watchdog handler as user-provided
	// code and show the frames on top of it.
	kroot(trampoline)();
    }

    virtual void dispose() {
	// Prevent deletion from CkAlarm::tick().
    }
};

class WindInterrupt : public CkInterrupt {

public:

    VOIDFUNCPTR routine;

    int parameter;

    WindInterrupt(ckvector_t vector,
		  cklevel_t level,
		  VOIDFUNCPTR routine,
		  int parameter,
		  const char *name,
		  WindNode *node);
};

#endif // !_WindNode_h
