/* -*- C++ -*-
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Interface to the WIND synchronization objects.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifndef _WindSynch_h
#define _WindSynch_h

#if defined(__GNUG__) && !defined(__OBSCAN__)
#pragma interface
#endif // __GNUG__ && !__OBSCAN__

#include "CkSynch.h"
#include "vxworks/vxWorks.h"

#ifdef IMPLEMENTING_SO
#undef IMPLEMENTING_SO
#endif // IMPLEMENTING_SO
#define IMPLEMENTING_SO IMPLEMENT_WIND
#include "sodecl.h"

class WindNode;
class WindTask;

class WindSynch : public CkSynchro {

protected:

    int options;

public:

    WindSynch(const char *name,
	      int options,
	      InsertMode mode,
	      CkStateExportObject *synchState,
	      WindNode *node);

    WindNode *getNode() {
	return (WindNode *)CkSynchro::getNode();
    }

    int testOption(int opt) const {
	return !!(options & opt);
    }

    void blockTask(WindTask *task,
		   ckticks_t timeout);

    void unblockTask(WindTask *task);

    virtual void forget(CkThread *thread);
};

enum WindSema4States {

    WindSema4IDLE =0,
    WindSema4PENDED =1,
    WindSema4POSTED =2
};

// These flags are cumulative with the WIND sema4 options
// defined in vxworks/semLib.h.
#define WIND_SEM_COUNTING 0x0
#define WIND_SEM_BINARY   0x80

class WindSema4 : public WindSynch {

protected:

    int count;

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[3];

public:

    WindSema4(int options,
	      int count,
	      WindNode *node);

    virtual ~WindSema4();

    int peek() {
    	return count;
    }

    int acquire(WindTask *task,
		ckticks_t timeout);

    int release();

    int broadcast();

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

enum WindMutexStates {

    WindMutexFREE =0,
    WindMutexLOCKED =1,
    WindMutexPENDED =2
};

class WindMutex : public WindSynch {

protected:

    WindTask *owner;

    int ownCount;

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[3];

public:

    WindMutex(int options,
	      WindNode *node);

    virtual ~WindMutex();

    int lock(WindTask *task,
	     ckticks_t timeout);

    int unlock(WindTask *task);

    int isLocked() const {
	return synchState->getState() == WindMutexLOCKED;
    }

    void safeSync(WindTask *waiter);

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

enum WindMsgQStates {

    WindMsgQEMPTY =0,
    WindMsgQPENDED =1,
    WindMsgQPOSTED =2,
    WindMsgQFULL =3
};

// This flag is cumulative with the WIND msgQ options
// defined in vxworks/msgQLib.h.
#define WIND_MSG_PRIVATE 0x80

struct WindMessage {

    int flags;
    unsigned length;
    ITime time;
    struct WindMessage *link;

    // Start of data area: must be last member.
    char data[1];
};

class WindQueue {

protected:

    WindMessage *qhead,
	**qtail;

    int count;

public:

    WindMessage *pop() {

	WindMessage *msg = qhead;

	if (msg)
	    {
	    qhead = msg->link;

	    if (--count == 0)
		qtail = &qhead;
	    }

	return msg;
    }

    void prepend(WindMessage *msg) {
	msg->link = qhead;
	qhead = msg;
	count++;
    }

    void append(WindMessage *msg) {
	*qtail = msg;
	qtail = &msg->link;
	msg->link = NULL;
	count++;
    }

    WindQueue() {
	qhead = NULL;
	qtail = &qhead;
	count = 0;
    }

    virtual ~WindQueue() {

	WindMessage *msg;

	while ((msg = pop()) != NULL)
	    delete[] (char *)msg;
    }

    int getCount() const {
	return count;
    }

    WindMessage *first() {
	return qhead;
    }
};

class WindMsgQ : public WindSynch {

    friend struct WindMessage;

protected:

    unsigned maxMsg;

    unsigned maxLength;

    int sendBlockCount;

    WindQueue freeq,
	inq;

    virtual void protoInit();

    virtual void dynamicExpose();

    virtual void dynamicConfigure(StProtoInfoMessage *mbuf);

    static const char *stateArray[4];

public:

    WindMsgQ(int options,
	     unsigned maxMsg,
	     unsigned maxLength,
	     WindNode *node);

    virtual ~WindMsgQ();

    int getMsgCount() const {
	return inq.getCount();
    }

    int getSendBlockCnt() const {
	return sendBlockCount;
    }

    int getRecvBlockCnt() const {
	return getCount() - sendBlockCount;
    }

    unsigned getMaxMsg() const {
	return maxMsg;
    }

    unsigned getMaxLength() const {
	return maxLength;
    }

    int send(WindTask *sender,
	     void *message,
	     unsigned length,
	     int options,
	     ckticks_t timeout);

    int receive(WindTask *receiver,
		ckticks_t timeout);

    void dispose(WindMessage *msg) {
	if (msg->flags & WIND_MSG_PRIVATE)
	    delete[] (char *)msg;
	else
	    freeq.append(msg);
    }

    virtual const char *getTypedName();

    virtual void forget(CkThread *thread);
};

#endif // !_WindSynch_h
