/*
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: The WIND kernel native API this model exports to applications.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#include <config.h>
#include <stdlib.h>
#include <string.h>
#include "CkManager.h"
#include "WindNode.h"
#include "WindTrace.h"

#define GET_ERRNO(cond,evar) \
          (evar = (cond) ? WindRunningNode->findCurrentRunnable()->getErrno() : 0)

BOOL intContext (void)

{
    int rc;

    CkRetInt rv(&rc);

    CkTracePoint t(0,NULL,&rv,NULL);

    rc = !!WindRunningNode->getIsrCount();

    t.kdoor(emit)();

    return rc;
}

int intCount (void)

{
    int rc;

    CkRetInt rv(&rc);

    CkTracePoint t(1,NULL,&rv,NULL);

    rc = WindRunningNode->getIsrCount();

    t.kdoor(emit)();

    return rc;
}

int intLevelSet (int level)

{
    int plevel;

    CkInInt p1("level",level);
    CkRetInt rv(&plevel);

    CkTracePoint t(2,NULL,&p1,&rv,NULL);

    if (level < 0)
	CkManager::This->fatal("intrLevelSet(level=%d): invalid level",level);

    if ((cklevel_t)level > WindRunningNode->getMaxIntrLevel())
	level = (int)WindRunningNode->getMaxIntrLevel();

    plevel = WindRunningNode->kdoor(imask)(level);

    t.kdoor(emit)();

    return plevel;
}

int intLock (void)

{
    int lockKey;

    CkRetInt rv(&lockKey);

    CkTracePoint t(3,NULL,&rv,NULL);

    lockKey = WindRunningNode->kdoor(imask)(WindRunningNode->getMaxIntrLevel());

    t.kdoor(emit)();

    return lockKey;
}

void intUnlock (int lockKey)

{
    CkInInt p1("lockKey",lockKey);

    CkTracePoint t(4,NULL,&p1,NULL);

    if (lockKey < 0 ||
	(cklevel_t)lockKey > WindRunningNode->getMaxIntrLevel())
	CkManager::This->fatal("intUnlock(lockKey=%d): invalid lock key",lockKey);

    WindRunningNode->kdoor(imask)(lockKey);

    t.kdoor(emit)();
}

STATUS kernelTimeSlice (int ticks)

{
    STATUS rc;
    int err;

    CkInInt p1("ticks",ticks);
    CkRetInt rv(&rc);

    CkTracePoint t(5,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(kernelTimeSlice)(ticks);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

const char *kernelVersion (void)

{
    const char *version;

    CkRetString rv(&version);

    CkTracePoint t(6,NULL,&rv,NULL);

    version = "not exactly WIND for CarbonKernel - version " WIND_MODEL_VERSION;

    return version;
}

MSG_Q_ID msgQCreate (int maxMsg,
		     unsigned maxLength,
		     int options)
{
    ckhandle_t qid;
    int err;

    CkInInt p1("maxMsg",maxMsg);
    CkInUInt p2("maxLength",maxLength);
    CkInFlags p3("options",options);
    CkRetHDL rv(&qid);

    CkTracePoint t(7,&err,&p1,&p2,&p3,&rv,NULL);

    qid = WindRunningNode->kdoor(msgQCreate)(maxMsg,
					     maxLength,
					     options);
    GET_ERRNO(qid == 0,err);

    t.kdoor(emit)();

    return qid;
}

STATUS msgQDelete (MSG_Q_ID qid)

{
    STATUS rc;
    int err;

    CkInHDL p1("qid",qid);
    CkRetInt rv(&rc);

    CkTracePoint t(8,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(msgQDelete)(qid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

int msgQNumMsgs (MSG_Q_ID qid)

{
    STATUS rc;
    int err;

    CkInHDL p1("qid",qid);
    CkRetInt rv(&rc);

    CkTracePoint t(9,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(msgQNumMsgs)(qid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

int msgQReceive (MSG_Q_ID qid,
		 char *buffer,
		 unsigned nBytes,
		 int ticks)
{
    STATUS rc;
    int err;

    CkInHDL p1("qid",qid);
    CkInAddr p2("buffer",buffer);
    CkInUInt p3("nBytes",nBytes);
    CkInInt p4("timeout",ticks);
    CkRetInt rv(&rc);

    CkTracePoint t(10,&err,&p1,&p2,&p3,&p4,&rv,NULL);

    rc = WindRunningNode->kdoor(msgQReceive)(qid,buffer,nBytes,ticks);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS msgQSend (MSG_Q_ID qid,
		 char *buffer,
		 unsigned nBytes,
		 int ticks,
		 int priority)
{
    STATUS rc;
    int err;

    CkInHDL p1("qid",qid);
    CkInAddr p2("buffer",buffer);
    CkInUInt p3("nBytes",nBytes);
    CkInInt p4("timeout",ticks);
    CkInInt p5("priority",priority);
    CkRetInt rv(&rc);

    CkTracePoint t(11,&err,&p1,&p2,&p3,&p4,&p5,&rv,NULL);

    rc = WindRunningNode->kdoor(msgQSend)(qid,buffer,nBytes,priority,ticks);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

SEM_ID semBCreate (int options, SEM_B_STATE initialState)

{
    ckhandle_t semid;
    int err;

    CkInFlags p1("options",options);
    CkInInt p2("initialState",initialState);
    CkRetHDL rv(&semid);

    CkTracePoint t(12,&err,&p1,&p2,&rv,NULL);

    semid = WindRunningNode->kdoor(semBCreate)(options,
					       initialState);
    GET_ERRNO(semid == 0,err);

    t.kdoor(emit)();

    return semid;
}

SEM_ID semCCreate (int options, int initialCount)

{
    ckhandle_t semid;
    int err;

    CkInFlags p1("options",options);
    CkInInt p2("initialCount",initialCount);
    CkRetHDL rv(&semid);

    CkTracePoint t(13,&err,&p1,&p2,&rv,NULL);

    semid = WindRunningNode->kdoor(semCCreate)(options,
					       initialCount);
    GET_ERRNO(semid == 0,err);

    t.kdoor(emit)();

    return semid;
}

SEM_ID semMCreate (int options)

{
    ckhandle_t semid;
    int err;

    CkInFlags p1("options",options);
    CkRetHDL rv(&semid);

    CkTracePoint t(14,&err,&p1,&rv,NULL);

    semid = WindRunningNode->kdoor(semMCreate)(options);

    GET_ERRNO(semid == 0,err);

    t.kdoor(emit)();

    return semid;
}

STATUS semDelete (SEM_ID semid)

{
    STATUS rc;
    int err;

    CkInHDL p1("semid",semid);
    CkRetInt rv(&rc);

    CkTracePoint t(15,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(semDelete)(semid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS semFlush (SEM_ID semid)

{
    STATUS rc;
    int err;

    CkInHDL p1("semid",semid);
    CkRetInt rv(&rc);

    CkTracePoint t(16,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(semFlush)(semid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS semGive (SEM_ID semid)

{
    STATUS rc;
    int err;

    CkInHDL p1("semid",semid);
    CkRetInt rv(&rc);

    CkTracePoint t(17,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(semGive)(semid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS semTake (SEM_ID semid, int ticks)

{
    STATUS rc;
    int err;

    CkInHDL p1("semid",semid);
    CkInInt p2("timeout",ticks);
    CkRetInt rv(&rc);

    CkTracePoint t(18,&err,&p1,&p2,&rv,NULL);

    rc = WindRunningNode->kdoor(semTake)(semid,ticks);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

void sysClkDisable (void)

{
    CkTracePoint t(19,NULL,NULL);

    WindRunningNode->rtcDisable();

    t.kdoor(emit)();
}

void sysClkEnable (void)

{
    CkTracePoint t(20,NULL,NULL);

    WindRunningNode->rtcEnable();

    t.kdoor(emit)();
}

int sysClkRateGet (void)

{
    int tps;

    CkRetInt rv(&tps);

    CkTracePoint t(21,NULL,&rv,NULL);

    tps = (int)WindRunningNode->getTicks2Sec();

    t.kdoor(emit)();

    return tps;
}

STATUS taskActivate (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(22,&err,&p1,&rv,NULL);

    if (!pTcb ||!TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskActivate(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskActivate)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskDelay (int ticks)

{
    STATUS rc;
    int err;

    CkInInt p1("ticks",ticks);
    CkRetInt rv(&rc);

    CkTracePoint t(23,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(taskDelay)(ticks);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskDelete (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(24,&err,&p1,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskDelete(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskDelete)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskDeleteForce (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(25,&err,&p1,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskDeleteForce(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskDeleteForce)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

int taskIdSelf (void)

{
    void *tcbAddr;
    int err;

    CkRetAddr rv(&tcbAddr);

    CkTracePoint t(26,&err,&rv,NULL);

    tcbAddr = WindRunningNode->kdoor(taskIdSelf)();

    GET_ERRNO(tcbAddr == NULL,err);

    t.kdoor(emit)();

    return (wind_tid_t)tcbAddr;
}

STATUS taskIdVerify (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(27,&err,&p1,&rv,NULL);

    // Don't even try to check for validity if the TCB pointer
    // is NULL or invalid.

    if (!pTcb || !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	{
	WindRunningNode->setContextErrno(S_objLib_OBJ_ID_ERROR);
	rc = ERROR;
	}
    else
	rc = WindRunningNode->kdoor(taskIdVerify)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskInit (WIND_TCB *pTcb,
		 const char *name,
		 int priority,
		 int options,
		 char *stackBase,
		 int stacksize,
		 FUNCPTR entryPt,
		 int arg1,
		 int arg2,
		 int arg3,
		 int arg4,
		 int arg5,
		 int arg6,
		 int arg7,
		 int arg8,
		 int arg9,
		 int arg10)
{
    STATUS rc;
    int err;

    CkInAddr p1("pTcb",pTcb);
    CkInString p2("name",name);
    CkInInt p3("priority",priority);
    CkInFlags p4("options",options);
    CkInAddr p5("stackBase",stackBase);
    CkInInt p6("stacksize",stacksize);
    CkInAddr p7("entryPt",(void *)entryPt);
    CkInInt p8("arg1",arg1);
    CkInInt p9("arg2",arg2);
    CkInInt p10("arg3",arg3);
    CkInInt p11("arg4",arg4);
    CkInInt p12("arg5",arg5);
    CkInInt p13("arg6",arg6);
    CkInInt p14("arg7",arg7);
    CkInInt p15("arg8",arg8);
    CkInInt p16("arg9",arg9);
    CkInInt p17("arg10",arg10);
    CkRetInt rv(&rc);

    CkTracePoint t(28,&err,&p1,&p2,&p3,&p4,&p5,&p6,&p7,&p8,&p9,
		   &p10,&p11,&p12,&p13,&p14,&p15,&p16,&p17,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskInit(pTcb=%p): invalid memory",pTcb);

    if (name && !TEST_PARAM(name,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskInit(name=%p): invalid memory",name);

    int args[10] = { arg1, arg2, arg3, arg4, arg5,
		     arg6, arg7, arg8, arg9, arg10 };

    rc = WindRunningNode->kdoor(taskInit) (pTcb,
					   name,
					   priority,
					   options,
					   stacksize,
					   entryPt,
					   args,
					   NULL);
    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskLock (void)

{
    STATUS rc;
    int err;

    CkRetInt rv(&rc);

    CkTracePoint t(29,&err,&rv,NULL);

    rc = WindRunningNode->kdoor(taskLock)();

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskPriorityGet (int tid, int *pPriority)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkInAddr p2("pPriority",pPriority);
    CkRetInt rv(&rc);

    CkTracePoint t(30,&err,&p1,&p2,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskPriorityGet(tid=%p): invalid tid",
			       pTcb);

    if (!pPriority || !TEST_PARAM(pPriority,sizeof(*pPriority),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskPriorityGet(pPriority=%p): invalid memory",
			       pPriority);

    rc = WindRunningNode->kdoor(taskPriorityGet)(pTcb,pPriority);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskPrioritySet (int tid, int priority)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkInInt p2("priority",priority);
    CkRetInt rv(&rc);

    CkTracePoint t(31,&err,&p1,&p2,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskPrioritySet(tid=%p): invalid tid",
			       pTcb);

    rc = WindRunningNode->kdoor(taskPrioritySet)(pTcb,priority);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskRestart (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(32,&err,&p1,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskRestart(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskRestart)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskResume (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(33,&err,&p1,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskResume(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskResume)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskSafe (void)

{
    STATUS rc;
    int err;

    CkRetInt rv(&rc);

    CkTracePoint t(34,&err,&rv,NULL);

    rc = WindRunningNode->kdoor(taskSafe)();

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

int taskSpawn (const char *name,
	       int priority,
	       int options,
	       int stacksize,
	       FUNCPTR entryPt,
	       int arg1,
	       int arg2,
	       int arg3,
	       int arg4,
	       int arg5,
	       int arg6,
	       int arg7,
	       int arg8,
	       int arg9,
	       int arg10)
{
    void *tcbAddr;
    int err;

    CkInString p1("name",name);
    CkInInt p2("priority",priority);
    CkInFlags p3("options",options);
    CkInInt p4("stacksize",stacksize);
    CkInAddr p5("entryPt",(void *)entryPt);
    CkInInt p6("arg1",arg1);
    CkInInt p7("arg2",arg2);
    CkInInt p8("arg3",arg3);
    CkInInt p9("arg4",arg4);
    CkInInt p10("arg5",arg5);
    CkInInt p11("arg6",arg6);
    CkInInt p12("arg7",arg7);
    CkInInt p13("arg8",arg8);
    CkInInt p14("arg9",arg9);
    CkInInt p15("arg10",arg10);
    CkRetAddr rv(&tcbAddr);

    CkTracePoint t(35,&err,&p1,&p2,&p3,&p4,&p5,&p6,&p7,&p8,
		   &p9,&p10,&p11,&p12,&p13,&p14,&p15,&rv,NULL);

    if (name && !TEST_PARAM(name,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskSpawn(name=%p): invalid memory",name);

    int args[10] = { arg1, arg2, arg3, arg4, arg5,
		     arg6, arg7, arg8, arg9, arg10 };

    tcbAddr = WindRunningNode->kdoor(taskSpawn)(name,
						priority,
						options,
						stacksize,
						entryPt,
						args);
    GET_ERRNO(tcbAddr == NULL,err);

    t.kdoor(emit)();

    return (wind_tid_t)tcbAddr;
}

STATUS taskSuspend (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    STATUS rc;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(36,&err,&p1,&rv,NULL);

    if (pTcb && !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskSuspend(tid=%p): invalid tid",pTcb);

    rc = WindRunningNode->kdoor(taskSuspend)(pTcb);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

WIND_TCB *taskTcb (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    int err;

    CkInAddr p1("tid",pTcb);
    CkRetAddr rv((void **)&pTcb);

    CkTracePoint t(37,&err,&p1,&rv,NULL);

    if (!pTcb || !TEST_PARAM(pTcb,sizeof(*pTcb),MEMADDR_WRITABLE))
	CkManager::This->fatal("taskTcb(tid=%p): invalid tid",pTcb);

    pTcb = WindRunningNode->kdoor(taskTcb)(pTcb);

    GET_ERRNO(pTcb == NULL,err);

    t.kdoor(emit)();

    return pTcb;
}

STATUS taskUnlock (void)

{
    STATUS rc;
    int err;

    CkRetInt rv(&rc);

    CkTracePoint t(38,&err,&rv,NULL);

    rc = WindRunningNode->kdoor(taskUnlock)();

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskUnsafe (void)

{
    STATUS rc;
    int err;

    CkRetInt rv(&rc);

    CkTracePoint t(39,&err,&rv,NULL);

    rc = WindRunningNode->kdoor(taskUnsafe)();

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

void taskExit__ (int status)

{
    CkInInt p1("status",status);

    CkTracePoint t(40,NULL,&p1,NULL);

    WindRunningNode->kdoor(taskExit)(status);

    t.kdoor(emit)();
}

STATUS taskCreateHookAdd (FUNCPTR createHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("createHook",(void *)createHook);
    CkRetInt rv(&rc);

    CkTracePoint t(41,&err,&p1,&rv,NULL);

    if (!createHook || !TEST_PARAM(createHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskCreateHookAdd(createHook=%p): invalid hook",
			       createHook);

    rc = WindRunningNode->kdoor(taskCreateHookAdd)(createHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskCreateHookDelete (FUNCPTR createHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("createHook",(void *)createHook);
    CkRetInt rv(&rc);

    CkTracePoint t(42,&err,&p1,&rv,NULL);

    if (!createHook || !TEST_PARAM(createHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskCreateHookDelete(createHook=%p): invalid hook",
			       createHook);

    rc = WindRunningNode->kdoor(taskCreateHookDelete)(createHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskDeleteHookAdd (FUNCPTR deleteHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("deleteHook",(void *)deleteHook);
    CkRetInt rv(&rc);

    CkTracePoint t(43,&err,&p1,&rv,NULL);

    if (!deleteHook || !TEST_PARAM(deleteHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskDeleteHookAdd(deleteHook=%p): invalid hook",
			       deleteHook);

    rc = WindRunningNode->kdoor(taskDeleteHookAdd)(deleteHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskDeleteHookDelete (FUNCPTR deleteHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("deleteHook",(void *)deleteHook);
    CkRetInt rv(&rc);

    CkTracePoint t(44,&err,&p1,&rv,NULL);

    if (!deleteHook || !TEST_PARAM(deleteHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskDeleteHookDelete(deleteHook=%p): invalid hook",
			       deleteHook);

    rc = WindRunningNode->kdoor(taskDeleteHookDelete)(deleteHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskSwitchHookAdd (FUNCPTR switchHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("switchHook",(void *)switchHook);
    CkRetInt rv(&rc);

    CkTracePoint t(45,&err,&p1,&rv,NULL);

    if (!switchHook || !TEST_PARAM(switchHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskSwitchHookAdd(switchHook=%p): invalid hook",
			       switchHook);

    rc = WindRunningNode->kdoor(taskSwitchHookAdd)(switchHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS taskSwitchHookDelete (FUNCPTR switchHook)

{
    STATUS rc;
    int err;

    CkInAddr p1("switchHook",(void *)switchHook);
    CkRetInt rv(&rc);

    CkTracePoint t(46,&err,&p1,&rv,NULL);

    if (!switchHook || !TEST_PARAM(switchHook,sizeof(char),MEMADDR_READABLE))
	CkManager::This->fatal("taskSwitchHookDelete(switchHook=%p): invalid hook",
			       switchHook);

    rc = WindRunningNode->kdoor(taskSwitchHookDelete)(switchHook);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

int taskIdDefault (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    void *dflTcb;

    CkInAddr p1("tid",pTcb);
    CkRetAddr rv(&dflTcb);

    CkTracePoint t(47,NULL,&p1,&rv,NULL);

    dflTcb = WindRunningNode->kdoor(taskIdDefault)(pTcb);

    t.kdoor(emit)();

    return (wind_tid_t)dflTcb;
}

BOOL taskIsReady (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    int rc;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(48,NULL,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(taskIsReady)(pTcb);

    t.kdoor(emit)();

    return rc;
}

BOOL taskIsSuspended (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    int rc;

    CkInAddr p1("tid",pTcb);
    CkRetInt rv(&rc);

    CkTracePoint t(49,NULL,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(taskIsSuspended)(pTcb);

    t.kdoor(emit)();

    return rc;
}

const char *taskName (int tid)

{
    WIND_TCB *pTcb = (WIND_TCB *)tid;
    const char *name;

    CkInAddr p1("tid",pTcb);
    CkRetString rv(&name);

    CkTracePoint t(50,NULL,&p1,&rv,NULL);

    name = WindRunningNode->kdoor(taskName)(pTcb);

    t.kdoor(emit)();

    return name;
}

int taskNameToId (const char *name)

{
    void *pTcb;

    CkInString p1("name",name);
    CkRetAddr rv(&pTcb);

    CkTracePoint t(51,NULL,&p1,&rv,NULL);

    if (!name || !TEST_PARAM(name,sizeof(*name),MEMADDR_READABLE))
	CkManager::This->fatal("taskName(name=%p): invalid memory",name);

    pTcb = WindRunningNode->kdoor(taskNameToId)(name);

    t.kdoor(emit)();

    return (wind_tid_t)pTcb;
}

u_long tickGet (void)

{
    u_long ticks;

    CkRetULong rv(&ticks);

    CkTracePoint t(52,NULL,&rv,NULL);

    ticks = WindRunningNode->kdoor(tickGet)();

    t.kdoor(emit)();

    return ticks;
}

void tickSet (u_long ticks)

{
    CkInULong p1("ticks",ticks);

    CkTracePoint t(53,NULL,&p1,NULL);

    WindRunningNode->kdoor(tickSet)(ticks);

    t.kdoor(emit)();
}

void tickAnnounce (int) {

    // Obviously no trace point for this one (!)
    WindRunningNode->kdoor(tick)();
}

STATUS wdCancel (WDOG_ID wid)

{
    STATUS rc;
    int err;

    CkInHDL p1("wid",wid);
    CkRetInt rv(&rc);

    CkTracePoint t(54,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(wdCancel)(wid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

WDOG_ID wdCreate (void)

{
    ckhandle_t wid;
    int err;

    CkRetHDL rv(&wid);

    CkTracePoint t(55,&err,NULL,&rv,NULL);

    wid = WindRunningNode->kdoor(wdCreate)();

    GET_ERRNO(wid == 0,err);

    t.kdoor(emit)();

    return wid;
}

STATUS wdDelete (WDOG_ID wid)

{
    STATUS rc;
    int err;

    CkInHDL p1("wid",wid);
    CkRetInt rv(&rc);

    CkTracePoint t(56,&err,&p1,&rv,NULL);

    rc = WindRunningNode->kdoor(wdDelete)(wid);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}

STATUS wdStart (WDOG_ID wid, int ticks,	FUNCPTR pRoutine, int param)

{
    STATUS rc;
    int err;

    CkInHDL p1("wid",wid);
    CkInInt p2("delay",ticks);
    CkInAddr p3("pRoutine",(void *)pRoutine);
    CkInInt p4("param",param);
    CkRetInt rv(&rc);

    CkTracePoint t(57,&err,&p1,&p2,&p3,&p4,&rv,NULL);

    rc = WindRunningNode->kdoor(wdStart)(wid,ticks,pRoutine,param);

    GET_ERRNO(rc == ERROR,err);

    t.kdoor(emit)();

    return rc;
}
