/*
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the WIND node configuration classes.
 * This code is based on the CfClass package.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include "WindConfig.h"
#include "WindNode.h"

BEGIN_RINFO
RDEFINE("VxWorks","VxWorks target",CKDB_OSNODE,WindConfig,CF_NONE,0,7)
END_RINFO

const struct CkInodeType WindInodeTypes[] = {

    { "task", WIND_TASK_INODE },
    { "mq", WIND_MQ_INODE },
    { "sem", WIND_SEMA4_INODE },
    { "mutex", WIND_MUTEX_INODE },
    { "watchdog", WIND_WD_INODE },
    { NULL, -1 }
};

WindConfig::WindConfig (const char *_instanceName,
			CfContext *_context) :
    CfClass(_instanceName,_context)
{
    // WIND kernel configuration group
    
    HandlerAttributePointer startAddrPtr =
	new HandlerAttribute("Entry point",this,"WIND");
    startAddrPtr->setValue("rootTask");

    // VRTOS configuration group

    ScaleAttributePointer warpPtr =
	new ScaleAttribute("Target Warp",this,0.0,10.0,0.1,"CarbonKernel");
    warpPtr->setValue(2.0);

    IntegerAttributePointer rtcVectorPtr =
	new IntegerAttribute("RTC interrupt vector",this,0,CK_VECTOR_UMAX,"CarbonKernel");
    rtcVectorPtr->setValue(0);

    IntegerAttributePointer rtcLevelPtr =
	new IntegerAttribute("RTC interrupt level",this,1,32,"CarbonKernel");
    rtcLevelPtr->setValue(1);

    HandlerAttributePointer rtcHandlerPtr =
	new HandlerAttribute("RTC service routine",this,"CarbonKernel");
    rtcHandlerPtr->setValue("tickAnnounce");

    RealAttributePointer rtcFreqPtr =
	new RealAttribute("RTC Frequency (Hz)",this,1,infinity(),"CarbonKernel");
    rtcFreqPtr->setValue(100.0);

    ResourceAttributePointer interruptPtr =
	new ResourceAttribute("Interrupts",this,CKDB_INTR,"CarbonKernel");

    ResourceAttributePointer panelPtr =
	new ResourceAttribute("Control panels",this,CKDB_PANEL,"CarbonKernel");
}

WindConfig::WindConfig (DBapl *apl) :
    CfClass(apl) {
}

WindConfig::WindConfig (WindConfig& src) :
    CfClass(src) {
}

void WindConfig::instantiate (CfContext *context)

{
    HandlerAttributePointer startAddrPtr(get("Entry point"));
    ScaleAttributePointer warpPtr(get("Target Warp"));
    HandlerAttributePointer rtcVectorPtr(get("RTC interrupt vector"));
    IntegerAttributePointer rtcLevelPtr(get("RTC interrupt level"));
    HandlerAttributePointer rtcHandlerPtr(get("RTC service routine"));
    RealAttributePointer rtcFreqPtr(get("RTC Frequency (Hz)"));
    ResourceAttributePointer interruptPtr(get("Interrupts"));
    ResourceAttributePointer panelPtr(get("Control panels"));

    // Console configuration block
    
    CkConsoleParms cparms;
    cparms.iclistMax = CK_CLIST_INFINITE;
    cparms.oclistMax = CK_CLIST_INFINITE;

    // VRTOS configuration block
    
    CkNodeParms nparms;

    nparms.nodeSerial = getSerialNumber(context);
    nparms.warp = warpPtr->getValue();
    nparms.priMode = PRUP; // i.e. The lower the value, the higher the priority
    nparms.maxIntrLevel = 32;
    nparms.rtcIntrVector = (ckvector_t)rtcVectorPtr->getValue();
    nparms.rtcIntrLevel = (cklevel_t)rtcLevelPtr->getValue();
    nparms.rtcIntrHandler = rtcHandlerPtr->resolveAddrIfAny(context);
    nparms.rtcHz = rtcFreqPtr->getValue();
    nparms.cparms = cparms;
    nparms.inodeTypes = WindInodeTypes;

    // Initialize the WIND node using current configuration.

    WindNode *node = new WindNode(getSerialName(context),&nparms);

    node->startAddr = startAddrPtr->resolveAddr(context);
    node->customOptions = 0;	// None defined as of now.
    
    CfContextList subContextList;

    if (interruptPtr.valid())
	{
	interruptPtr->instantiate(context,subContextList,node);
	subContextList.destroy();
	}
    
    if (panelPtr.valid())
	{
	panelPtr->instantiate(context,subContextList,node);
	subContextList.destroy();
	}
}
