/*
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the WIND target abstraction.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation "WindNode.h"
#pragma implementation "WindProto.h"
#endif // __GNUG__
#include <config.h>
#include <sys/types.h>
#include <values.h>
#include <math.h>
#include <string.h>
#include <memory.h>
#include <errno.h>
#include "toolshop.h"
#include "CkManager.h"
#include "WindNode.h"
#include "WindProto.h"
#include "WindTask.h"
#include "WindTrace.h"
#include "WindSynch.h"

WindNode::WindNode (const char *_nodeName,
		    CkNodeParms *_nparms) :
    CkNode("VxWorks",
	   _nodeName,
	   WIND_MAGIC,
	   _nparms)
{
    schedLocks = 0;

    // Configure and start the real-time clock. A value of zero for the
    // ticks2sec configuration value disables the RTC for the node.
    // A null handler also prevents the RTC installation.

    if (getTicks2Sec() > 0 && _nparms->rtcIntrHandler != NULL)
	{
	// Create the interrupt object underlying the system timer.

	rtc = createInterrupt(_nparms->rtcIntrVector,
			      _nparms->rtcIntrLevel,
			      (ckisr_t *)_nparms->rtcIntrHandler,
			      NULL,
			      0,
			      CK_RTC_NAME);

	rtcProgram(getTicks2Sec());
	}

    setTraceInterface(new WindTraceInterface(this));
}

CkInterrupt *WindNode::createInterrupt (ckvector_t vector,
					cklevel_t level,
					ckisr_t *isr,
					ckdsr_t *dsr,
					void *cookie,
					const char *symbolicName)
{
    return new WindInterrupt(vector,
			     level,
			     (VOIDFUNCPTR)isr,
			     (int)cookie,
			     symbolicName,
			     this);
}

// warmInit -- proceeds to the warm initialization chores of a WIND node.
// This currently means spawning the root task. The scheduling is
// automatically started by the VRTOS after this method has returned.

void WindNode::kinit(warmInit) (void)

{
    CkNode::kinit(warmInit)();

    int args[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    WindTask *rootTask = new WindTask(NULL,
				      "Root",
				      0,
				      (FUNCPTR)startAddr,
				      0,
				      args,
				      this,
				      32768);
    declareThread(rootTask);
    startThread(rootTask);
}

void WindNode::protoInit (void)

{
    WindNodeExportMessage nem(this);
    protoExport(&nem,sizeof(nem));
}

ckticks_t WindNode::rrGetCredit (void) {

    return nslice;
}

void (*WindNode::getRootEntry (void)) (...) {

    return (void (*)(...))startAddr;
}

void WindNode::nameObject (CkObject *object, int type)

{
    int seqno = allocID();

    if (seqno < 0)
	// Oops, no more unique identifier available.
	CkManager::This->fatal("cannot generate unique identifiers");

    CkId id(getNid(),type,0,seqno);
    mapObject(id,type,CkLocalDomain,object);
    object->setCkId(id);
    object->setCkExtId(seqno);
    tellNameServer(id,CK_MSGPORT_NAMEOBJ);
}

void WindNode::nameObject (CkObject *object, int type, CkExtId uname)

{
    nameObject(object,type);
    object->setCkExtId(uname);
}

void WindNode::unnameObject (CkObject *object)

{
    unmapObject((CkHandle)object->getCkId(),
		object->getCkType(),
		CkLocalDomain);
    releaseID(object->getCkId().getSeq());
    tellNameServer(object->getCkId(),CK_MSGPORT_UNNAMEOBJ);
}

CkObject *WindNode::findActiveObject (CkId id) {

    return findObject((CkHandle)id,id.getType(),CkLocalDomain);
}

WindTask *WindNode::findTaskByTcb (WIND_TCB *pTcb)

{
    if (!pTcb)	// NULL tid means 'self'.
	return (WindTask *)getActiveThread();

    CkId threadid(pTcb->handle);
    WindTask *task = NULL;

    if (threadid.getType() == WIND_TASK_INODE)
	task = (WindTask *)findActiveObject(threadid);

    return task;
}

WindSema4 *WindNode::findSemByHandle (CkHandle handle)

{
    CkId semid(handle);
    WindSema4 *sema4 = NULL;

    if (semid.getType() == WIND_SEMA4_INODE)
	sema4 = (WindSema4 *)findActiveObject(semid);

    return sema4;
}

WindMutex *WindNode::findMutexByHandle (CkHandle handle)

{
    CkId mutexid(handle);
    WindMutex *mutex = NULL;

    if (mutexid.getType() == WIND_MUTEX_INODE)
	mutex = (WindMutex *)findActiveObject(mutexid);

    return mutex;
}

WindMsgQ *WindNode::findQueueByHandle (CkHandle handle)

{
    CkId qid(handle);
    WindMsgQ *mq = NULL;

    if (qid.getType() == WIND_MQ_INODE)
	mq = (WindMsgQ *)findActiveObject(qid);

    return mq;
}

WindWDog *WindNode::findWDogByHandle (CkHandle handle)

{
    CkId wid(handle);
    WindWDog *wd = NULL;

    if (wid.getType() == WIND_WD_INODE)
	wd = (WindWDog *)findActiveObject(wid);

    return wd;
}

// Issue a warning if an attempt is made to call a synchronous
// service on behalf of an asynchronous context. This method is
// a graceful replacement of CkNode::mustBeSynch() which merely
// aborts the simulation when a problem is detected.

STATUS WindNode::checkTaskContext (const char *syscall)

{
    if (!inTaskContext())
	{
	findCurrentRunnable()->setErrno(S_intLib_NOT_ISR_CALLABLE);
	CkManager::This->warning("'%s' is not callable on behalf of an ISR",syscall);
	return ERROR;
	}

    return OK;
}

// This method is needed for compatibility reasons since
// CkRunnable::setErrno() is not virtual in Ck 1.3.

void WindNode::setContextErrno (int errorNumber)

{
    if (inTaskContext())
	WindRunningTask->setErrno(errorNumber);
    else
	findCurrentRunnable()->setErrno(errorNumber);
}

int WindNode::dispatchSignals (WindTask *task)

{
    // FIXME: Not fully implemented
    checkSignals(task);

    return WIND_SIG_INTR;
}

void WindNode::kcout(fireCallout) (CkEventKind event, CkObject *object)

{
    CkNode::kcout(fireCallout)(event,object);

    WindEventHookList *hooks;

    switch (event)
	{
	case XTStartEvent:

	    hooks = &createHookList;
	    break;

	case XTSwitchEvent:
	    
	    hooks = &switchHookList;
	    break;

	case XTDeleteEvent:

	    hooks = &deleteHookList;
	    break;

	default:

	    hooks = NULL;
	    break;
	}

    if (!hooks)
	return;

    WindTask *task = (WindTask *)object; // Created/Incoming/Deleted
    void *arg = task->getTcb();

    for (WindEventHook *h = hooks->first();
	 h; h = (WindEventHook *)h->next())
	h->fn((int)arg);
}

STATUS WindNode::kdoor(kernelTimeSlice) (int ticks)

{
    if (checkTaskContext("kernelTimeSlice") == ERROR)
	return ERROR;

    nslice = ticks > 0 ? ticks : CK_INFINITE;

    if (nslice != CK_INFINITE)
	rrActivate();
    else
	rrDeactivate();

    return OK;
}

void WindNode::kdoor(tickSet) (ckticks_t ticks) {

    setNodeDate(ticks);
}

ckticks_t WindNode::kdoor(tickGet) (void) {

    return getNodeDate();
}

STATUS WindNode::kdoor(taskActivate) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskActivate") == ERROR)
	return ERROR;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    readyThread(task,CKSUSP);

    resched();

    return OK;
}

STATUS WindNode::kdoor(taskCreateHookAdd) (FUNCPTR createHook)

{
    WindEventHook *h = new WindEventHook(createHook);
    createHookList.append(h);

    return OK;
}

STATUS WindNode::kdoor(taskSwitchHookAdd) (FUNCPTR switchHook)

{
    WindEventHook *h = new WindEventHook(switchHook);
    switchHookList.append(h);

    return OK;
}

STATUS WindNode::kdoor(taskDeleteHookAdd) (FUNCPTR deleteHook)

{
    WindEventHook *h = new WindEventHook(deleteHook);
    deleteHookList.append(h);

    return OK;
}

STATUS WindNode::kdoor(taskCreateHookDelete) (FUNCPTR createHook)

{
    for (WindEventHook *h = createHookList.first();
	 h; h = (WindEventHook *)h->next())
	{
	if (h->fn == createHook)
	    {
	    createHookList.remove(h);
	    delete h;
	    return OK;
	    }
	}

    setContextErrno(S_taskLib_TASK_HOOK_NOT_FOUND);

    return ERROR;
}

STATUS WindNode::kdoor(taskSwitchHookDelete) (FUNCPTR switchHook)

{
    for (WindEventHook *h = switchHookList.first();
	 h; h = (WindEventHook *)h->next())
	{
	if (h->fn == switchHook)
	    {
	    switchHookList.remove(h);
	    delete h;
	    return OK;
	    }
	}

    setContextErrno(S_taskLib_TASK_HOOK_NOT_FOUND);

    return ERROR;
}

STATUS WindNode::kdoor(taskDeleteHookDelete) (FUNCPTR deleteHook)

{
    for (WindEventHook *h = deleteHookList.first();
	 h; h = (WindEventHook *)h->next())
	{
	if (h->fn == deleteHook)
	    {
	    deleteHookList.remove(h);
	    delete h;
	    return OK;
	    }
	}

    setContextErrno(S_taskLib_TASK_HOOK_NOT_FOUND);

    return ERROR;
}

STATUS WindNode::kdoor(taskDelay) (int ticks)

{
    if (checkTaskContext("taskDelay") == ERROR)
	return ERROR;

    if (ticks == NO_WAIT)
	// i.e. Round-robin in priority group.
	{
	readyThread(WindRunningTask);
	resched();
	}
    else
	{
	suspendThread(WindRunningTask,CKDELAY,ticks);

	resched();

	if (WindRunningTask->tstStatus(CKBREAK)) // Signal caught?
	    {
	    dispatchSignals(WindRunningTask);
	    setContextErrno(EINTR);

	    return ERROR;
	    }
	}

    return OK;
}

STATUS WindNode::kdoor(taskDelete) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskDelete") == ERROR)
	return ERROR;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    if (task != WindRunningTask)
	{
	CkId taskid = task->getCkId();

	task->safeSync(WindRunningTask);

	// Just in case the target task has self-deleted
	// inside its safe section: revalidate the task handle.

	if (!findActiveObject(taskid))
	    {
	    setContextErrno(S_objLib_OBJ_ID_ERROR);
	    return ERROR;
	    }
	}
	    
    deleteThread(task);

    resched();

    return OK;
}

STATUS WindNode::kdoor(taskDeleteForce) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskDeleteForce") == ERROR)
	return ERROR;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    deleteThread(task);

    resched();

    return OK;
}

void WindNode::kdoor(taskExit) (int status)

{
    mustBeSynch();

    WindTask *task = WindRunningTask;

    // Store the exit code in TCB to allow post-mortem processing
    // in Delete Hook.

    task->getTcb()->status = status;

    deleteThread(task);

    resched();
}

WIND_TCB *WindNode::kdoor(taskIdSelf) (void)

{
    if (checkTaskContext("taskIdSelf") == ERROR)
	return NULL;

    return WindRunningTask->getTcb();
}

STATUS WindNode::kdoor(taskIdVerify) (WIND_TCB *pTcb)

{
    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    return OK;
}

STATUS WindNode::kdoor(taskInit) (WIND_TCB *pTcb,
				  const char *suffix,
				  int priority,
				  int options,
				  int stackSize,
				  FUNCPTR entryPt,
				  int args[10],
				  WIND_TCB **ppTcb)
{
    CString name;

    if (checkTaskContext("taskInit") == ERROR)
	return ERROR;

    if (priority < 0 || priority > 255)
	{
	setContextErrno(S_taskLib_ILLEGAL_PRIORITY);
	return ERROR;
	}

    if (!suffix)
	name.format("t%u",++WindTask::seqGen);
    else
	name.format("t%s",suffix);

    WindTask *newTask = new WindTask(pTcb,
				     name,
				     priority,
				     entryPt,
				     options,
				     args,
				     this,
				     stackSize);

    // WIND tasks need to be started after creation using
    // taskActivate().

    declareThread(newTask);
    startThread(newTask);
    suspendThread(newTask,CKSUSP);

    // While dangerous - mainly for portability reasons (e.g. 64bit
    // arch), the equality tid == (int)&tcb is currently enforced
    // in this model since VxWorks 5.4 documentation explicitely
    // states it, and I have no doubt some application code may rely
    // on it... This also means that we'll have to keep our fake
    // WIND_TCB struct in sync with the current task status.
    // VxWorks 6.0 has introduced taskCreate() which returns an opaque
    // handle probably to get rid of this oddity.

    if (ppTcb)
	*ppTcb = newTask->getTcb();

    return OK;
}

STATUS WindNode::kdoor(taskLock) (void)

{
    if (checkTaskContext("taskLock") == ERROR)
	return ERROR;

    if (++schedLocks & 0xff000000)	// FIXME: Is there a maximum lock count?
	CkManager::This->fatal("taskLock(): scheduler overlocked");

    if (schedLocks == 1)
	WindRunningTask->lock();

    return OK;
}

STATUS WindNode::kdoor(taskPrioritySet) (WIND_TCB *pTcb,
					 int priority)
{
    if (priority < 0 || priority > 255)
	{
	setContextErrno(S_taskLib_ILLEGAL_PRIORITY);
	return ERROR;
	}

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    // Avoid manual round-robin if the priority doesn't change.

    if (priority != task->getPriNoPip())
	{
	reniceThread(task,priority);
	resched();
	}

    return OK;
}

STATUS WindNode::kdoor(taskPriorityGet) (WIND_TCB *pTcb,
					 int *pPriority)
{
    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    *pPriority = task->getPriNoPip();

    return OK;
}

STATUS WindNode::kdoor(taskRestart) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskRestart") == ERROR)
	return ERROR;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    if (task != WindRunningTask)
	{
	CkId taskid = task->getCkId();

	task->safeSync(WindRunningTask);

	// Just in case the target task has self-deleted
	// inside its safe section: revalidate the task handle.

	if (!findActiveObject(taskid))
	    {
	    setContextErrno(S_objLib_OBJ_ID_ERROR);
	    return ERROR;
	    }
	}

    task->restart();

    resched();

    return OK;
}

STATUS WindNode::kdoor(taskResume) (WIND_TCB *pTcb)

{
    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    if (getActiveThread() == task)
	CkManager::This->warning("taskResume(): self-resuming task ?!");
    else
	{
	readyThread(task,CKSUSP);
	resched();
	}

    return OK;
}

STATUS WindNode::kdoor(taskSafe) (void)

{
    if (checkTaskContext("taskSafe") == ERROR)
	return ERROR;

    WindRunningTask->enterSafeSection();

    // The rescheduling proc has already been called.

    return OK;
}

WIND_TCB *WindNode::kdoor(taskSpawn) (const char *suffix,
				      int priority,
				      int options,
				      int stackSize,
				      FUNCPTR entryPt,
				      int args[10])
{
    WIND_TCB *pTcb;

    if (checkTaskContext("taskSpawn") == ERROR)
	return NULL;

    if (kdoor(taskInit)(NULL,
			suffix,
			priority,
			options,
			stackSize,
			entryPt,
			args,
			&pTcb) != OK)
	return 0;

    WindTask *task = findTaskByTcb(pTcb);

    // Activate the new task immediately.

    readyThread(task,CKSUSP);

    resched();

    return pTcb;
}

STATUS WindNode::kdoor(taskSuspend) (WIND_TCB *pTcb)

{
    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    suspendThread(task,CKSUSP);

    resched();

    return OK;
}

WIND_TCB *WindNode::kdoor(taskTcb) (WIND_TCB *pTcb)

{
    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return NULL;
	}

    return pTcb;	// pTcb validity confirmed.
}

STATUS WindNode::kdoor(taskUnlock) (void)

{
    if (checkTaskContext("taskUnlock") == ERROR)
	return ERROR;

    if (!schedLocks)
	CkManager::This->warning("taskUnlock(): scheduler not locked");
    else if (--schedLocks == 0)
	{
	WindRunningTask->unlock();
	resched();
	}

    return OK;
}

STATUS WindNode::kdoor(taskUnsafe) (void)

{
    if (checkTaskContext("taskUnsafe") == ERROR)
	return ERROR;

    WindRunningTask->exitSafeSection();

    // The rescheduling proc has already been called.

    return OK;
}

CkHandle WindNode::kdoor(semBCreate) (int options,
				      SEM_B_STATE initialState)
{
    if (checkTaskContext("semBCreate") == ERROR)
	return CkNoId;

    if (options & ~SEM_Q_PRIORITY)
	{
	setContextErrno(S_semLib_INVALID_OPTION);
	return CkNoId;
	}

    if (initialState != SEM_EMPTY &&
	initialState != SEM_FULL)
	{
	setContextErrno(S_semLib_INVALID_STATE);
	return CkNoId;
	}

    WindSema4 *sema4 = new WindSema4(options|WIND_SEM_BINARY,
				     (int)initialState,
				     this);
    return sema4->getCkId();
}

CkHandle WindNode::kdoor(semCCreate) (int options,
				    int initialCount)
{
    if (checkTaskContext("semCCreate") == ERROR)
	return CkNoId;

    if (options & ~SEM_Q_PRIORITY)
	{
	setContextErrno(S_semLib_INVALID_OPTION);
	return CkNoId;
	}

    WindSema4 *sema4 = new WindSema4(options,initialCount,this);

    return sema4->getCkId();
}

CkHandle WindNode::kdoor(semMCreate) (int options)

{
    if (checkTaskContext("semMCreate") == ERROR)
	return CkNoId;

    if (options & ~SEM_OPTION_MASK)
	{
	setContextErrno(S_semLib_INVALID_OPTION);
	return CkNoId;
	}

    WindMutex *mutex = new WindMutex(options,this);

    return mutex->getCkId();
}

STATUS WindNode::kdoor(semDelete) (CkHandle handle)

{
    if (checkTaskContext("semDelete") == ERROR)
	return ERROR;

    WindMutex *mutex = findMutexByHandle(handle);

    if (mutex)
	{
	CkId mutexid = mutex->getCkId();

	mutex->safeSync(WindRunningTask);

	// Just in case the last owner has deleted the
	// mutex inside its safe section: revalidate the
	// mutex handle.

	if (!findActiveObject(mutexid))
	    {
	    setContextErrno(S_objLib_OBJ_ID_ERROR);
	    return ERROR;
	    }

	delete mutex;
	resched();

	return OK;
	}

    WindSema4 *sema4 = findSemByHandle(handle);
    
    if (sema4)
	{
	delete sema4;
	resched();

	return OK;
	}

    setContextErrno(S_objLib_OBJ_ID_ERROR);

    return ERROR;
}

STATUS WindNode::kdoor(semFlush) (CkHandle handle)

{
    if (checkTaskContext("semFlush") == ERROR)
	return ERROR;

    WindSema4 *sema4 = findSemByHandle(handle);
    
    if (sema4)
	{
	if (sema4->broadcast() == CK_SYNCH_RESCHED)
	    resched();

	return OK;
	}

    setContextErrno(findMutexByHandle(handle) ?
		    S_semLib_INVALID_OPERATION :
		    S_objLib_OBJ_ID_ERROR);

    return ERROR;
}

STATUS WindNode::kdoor(semTake) (CkHandle handle,
				 int ticks)
{
    ckticks_t timeout;

    if (checkTaskContext("semTake") == ERROR)
	return ERROR;

    if (ticks == NO_WAIT)
	timeout = CK_NONBLOCK;
    else if (ticks == WAIT_FOREVER)
	timeout = CK_INFINITE;
    else
	timeout = ticks;

restart:
	
    WindMutex *mutex = findMutexByHandle(handle);
    int s;

    if (mutex)
	s = mutex->lock(WindRunningTask,timeout);
    else
	{
	WindSema4 *sema4 = findSemByHandle(handle);

	if (sema4)
	    s = sema4->acquire(WindRunningTask,timeout);
	else
	    {
	    setContextErrno(S_objLib_OBJ_ID_ERROR);
	    return ERROR;
	    }
	}
    
    if (s == CK_SYNCH_WAIT)
	{
	resched();

	if (WindRunningTask->tstStatus(CKBREAK)) // Signal caught?
	    {
	    if (dispatchSignals(WindRunningTask) == WIND_SIG_RESTART)
		goto restart;

	    setContextErrno(EINTR);
	    return ERROR;
	    }

	if (WindRunningTask->tstStatus(CKTIMEO)) // Timeout elapsed?
	    {
	    setContextErrno(S_objLib_OBJ_TIMEOUT);
	    return ERROR;
	    }

	if (WindRunningTask->tstStatus(CKRMID)) // Semaphore deleted?
	    {
	    setContextErrno(S_objLib_OBJ_DELETED);
	    return ERROR;
	    }
	}
    else if (s == CK_SYNCH_NOWAIT)
	{
	setContextErrno(S_objLib_OBJ_UNAVAILABLE);
	return ERROR;
	}
    
    return OK;
}

STATUS WindNode::kdoor(semGive) (CkHandle handle)

{
    WindMutex *mutex = findMutexByHandle(handle);

    if (mutex)
	{
	if (checkTaskContext("semGive") == ERROR)
	    return ERROR;

	int s = mutex->unlock(WindRunningTask);

	if (s == CK_SYNCH_ERROR)
	    {
	    // Mutex unlocked or not owner.
	    setContextErrno(S_semLib_INVALID_OPERATION);
	    return ERROR;
	    }

	if (s == CK_SYNCH_RESCHED)
	    resched();

	return OK;
	}

    WindSema4 *sema4 = findSemByHandle(handle);

    if (sema4)
	{
	if (sema4->release() == CK_SYNCH_RESCHED)
	    resched();
	
	return OK;
	}

    setContextErrno(S_objLib_OBJ_ID_ERROR);

    return ERROR;
}

CkHandle WindNode::kdoor(msgQCreate) (int maxMsg,
				      unsigned maxLength,
				      int options)
{
    if (checkTaskContext("msgQCreate") == ERROR)
	return CkNoId;

    // FIXME: Are the two following sanity ckecks performed by VxWorks?

    if (options & ~MSG_Q_PRIORITY)
	{
	setContextErrno(S_msgQLib_INVALID_QUEUE_TYPE);
	return CkNoId;
	}

    if (maxMsg <= 0) // FIXME: Is null-queue allowed?
	{
	setContextErrno(S_msgQLib_INVALID_QUEUE_TYPE); // FIXME: error code
	return CkNoId;
	}

    if (maxLength == 0)
	{
	setContextErrno(S_msgQLib_INVALID_MSG_LENGTH);
	return CkNoId;
	}

    WindMsgQ *mq = new WindMsgQ(options,maxMsg,maxLength,this);

    return mq->getCkId();
}

STATUS WindNode::kdoor(msgQDelete) (CkHandle handle)

{
    if (checkTaskContext("msgQDelete") == ERROR)
	return ERROR;

    WindMsgQ *mq = findQueueByHandle(handle);
    
    if (mq)
	{
	delete mq;
	resched();

	return OK;
	}

    setContextErrno(S_objLib_OBJ_ID_ERROR);

    return ERROR;
}

STATUS WindNode::kdoor(msgQSend) (CkHandle handle,
				  char *buffer,
				  unsigned nBytes,
				  int priority,
				  int ticks)
{
    WindMsgQ *mq = findQueueByHandle(handle);
    
    if (!mq)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    if (priority & ~MSG_PRI_URGENT)
	{
	setContextErrno(S_msgQLib_INVALID_QUEUE_TYPE); // FIXME: error code
	return ERROR;
	}

    if (nBytes == 0)
	{
	setContextErrno(S_msgQLib_INVALID_MSG_LENGTH);
	return ERROR;
	}

    WindTask *sender = inTaskContext() ? WindRunningTask : NULL;
    ckticks_t timeout;

    if (ticks == NO_WAIT)
	timeout = CK_NONBLOCK;
    else
	{
	if (!inTaskContext())
	    {
	    setContextErrno(S_msgQLib_NON_ZERO_TIMEOUT_AT_INT_LEVEL);
	    return ERROR;
	    }

	if (ticks == WAIT_FOREVER)
	    timeout = CK_INFINITE;
	else
	    timeout = ticks;
	}

restart:

    int s = mq->send(sender,buffer,nBytes,priority,timeout);

    if (s == CK_SYNCH_RESCHED || s == CK_SYNCH_WAIT)
	{
	resched();

	if (s == CK_SYNCH_WAIT)
	    {
	    if (WindRunningTask->tstStatus(CKBREAK)) // Signal caught?
		{
		if (dispatchSignals(WindRunningTask) == WIND_SIG_RESTART)
		    goto restart;

		setContextErrno(EINTR);
		return ERROR;
		}

	    if (WindRunningTask->tstStatus(CKTIMEO)) // Timeout elapsed?
		{
		setContextErrno(S_objLib_OBJ_TIMEOUT);
		return ERROR;
		}

	    if (WindRunningTask->tstStatus(CKRMID)) // Message queue deleted?
		{
		setContextErrno(S_objLib_OBJ_DELETED);
		return ERROR;
		}
	    }
	}
    else if (s == CK_SYNCH_NOWAIT)
	{
	setContextErrno(S_objLib_OBJ_UNAVAILABLE);
	return ERROR;
	}

    return OK;
}

int WindNode::kdoor(msgQReceive) (CkHandle handle,
				  char *buffer,
				  unsigned nBytes,
				  int ticks)
{
    if (checkTaskContext("msgQReceive") == ERROR)
	return ERROR;

    WindMsgQ *mq = findQueueByHandle(handle);
    
    if (!mq)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    if (nBytes == 0)
	{
	setContextErrno(S_msgQLib_INVALID_MSG_LENGTH);
	return ERROR;
	}

    ckticks_t timeout;

    if (ticks == NO_WAIT)
	timeout = CK_NONBLOCK;
    else if (ticks == WAIT_FOREVER)
	timeout = CK_INFINITE;
    else
	timeout = ticks;

restart:

    int s = mq->receive(WindRunningTask,timeout);

    if (s == CK_SYNCH_RESCHED || s == CK_SYNCH_WAIT)
	{
	resched();

	if (s == CK_SYNCH_WAIT)
	    {
	    if (WindRunningTask->tstStatus(CKBREAK)) // Signal caught?
		{
		if (dispatchSignals(WindRunningTask) == WIND_SIG_RESTART)
		    goto restart;

		setContextErrno(EINTR);
		return ERROR;
		}

	    if (WindRunningTask->tstStatus(CKTIMEO)) // Timeout elapsed?
		{
		setContextErrno(S_objLib_OBJ_TIMEOUT);
		return ERROR;
		}

	    if (WindRunningTask->tstStatus(CKRMID)) // Message queue deleted?
		{
		setContextErrno(S_objLib_OBJ_DELETED);
		return ERROR;
		}
	    }
	}
    else if (s == CK_SYNCH_NOWAIT)
	{
	setContextErrno(S_objLib_OBJ_UNAVAILABLE);
	return ERROR;
	}

    WindMessage *msg = WindRunningTask->msgGet();
    int cBytes = 0;

    if (msg)
	{
	cBytes = Min(nBytes,msg->length);
	memcpy(buffer,msg->data,cBytes);
	mq->dispose(msg);
	}

    return cBytes;
}

int WindNode::kdoor(msgQNumMsgs) (CkHandle handle)

{
    // FIXME: is this service ISR callable?

    WindMsgQ *mq = findQueueByHandle(handle);
    
    if (!mq)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    return mq->getMsgCount();
}

CkHandle WindNode::kdoor(wdCreate) (void)

{
    if (checkTaskContext("wdCreate") == ERROR)
	return CkNoId;

    WindWDog *wd = new WindWDog(this);

    return wd->getCkId();
}

STATUS WindNode::kdoor(wdDelete) (CkHandle handle)

{
    if (checkTaskContext("wdDelete") == ERROR)
	return ERROR;

    WindWDog *wd = findWDogByHandle(handle);

    if (!wd)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    cancelAlarm(wd);
    delete wd;	// Because WindWDog::dispose() is a no-op

    return OK;
}

STATUS WindNode::kdoor(wdStart) (CkHandle handle,
				 int ticks,
				 FUNCPTR pRoutine,
				 int param)
{
    WindWDog *wd = findWDogByHandle(handle);

    if (!wd)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    wd->start(pRoutine,param,ticks);

    return OK;
}

STATUS WindNode::kdoor(wdCancel) (CkHandle handle)

{
    WindWDog *wd = findWDogByHandle(handle);

    if (!wd)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return ERROR;
	}

    wd->stop();

    return OK;
}

// WatchDog object -- May live after the creator task died.

WindWDog::WindWDog (WindNode *_node) :
    CkAlarm(_node) {
    _node->nameObject(this,WIND_WD_INODE);
}

WindWDog::~WindWDog () {
    getNode()->unnameObject(this);
}

void WindWDog::start (FUNCPTR _fn, int _arg, ckticks_t _ticks)

{
    fn = _fn;
    arg = _arg;
    set(_ticks,0,1);

    // CkAlarmManager::schedule() gracefully deals with
    // alarms which are already scheduled.

    getNode()->getAlarmManager()->schedule(this);
}

void WindWDog::stop (void) {

    // CkAlarmManager::forget() gracefully deals with
    // alarms which are not scheduled.

    getNode()->getAlarmManager()->forget(this);
}

WIND_TCB *WindNode::kdoor(taskIdDefault) (WIND_TCB *pTcb)

{
    WIND_TCB *dflTcb = NULL;

    if (defaultTask)
	dflTcb = defaultTask->getTcb();

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	CkManager::This->warning("taskIdDefault(tid=%p): attempt to set invalid default tid",
				 pTcb);
    else
	defaultTask = task;

    return dflTcb;
}

const char *WindNode::kdoor(taskName) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskName") == ERROR)
	return NULL;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return NULL;
	}

    return task->protoGetName();
}

WIND_TCB *WindNode::kdoor(taskNameToId) (const char *name)

{
    // At this point, name is valid.

    if (checkTaskContext("taskNameToId") == ERROR)
	return NULL;

    WindTaskIterator it(allWindTasks);
    WindTask *task;

    while ((task = it.next()) != NULL)
	{
	if ((*name == 't' && !strcmp(task->protoGetName(),name)) ||
	    !strcmp(task->protoGetName() + 1,name))
	    return task->getTcb();
	}

    setContextErrno(S_taskLib_NAME_NOT_FOUND);

    return NULL;
}

int WindNode::kdoor(taskIsReady) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskIsReady") == ERROR)
	return 0;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return 0;
	}

    return !task->tstStatus(CK_WAIT_BITS);
}

int WindNode::kdoor (taskIsSuspended) (WIND_TCB *pTcb)

{
    if (checkTaskContext("taskIsSuspended") == ERROR)
	return 0;

    WindTask *task = findTaskByTcb(pTcb);

    if (!task)
	{
	setContextErrno(S_objLib_OBJ_ID_ERROR);
	return 0;
	}

    return !!task->tstStatus(CK_WAIT_BITS);
}

// Export Protocol object

WindNodeExportMessage::WindNodeExportMessage (WindNode *node) :
    CkNodeExportMessage(node->getName())
{
    nid = node->getNid();
    scopy(osName,node->getOS(),sizeof(osName) - 1);
    strcpy(threadTypeName,"Task");
    asciiXID = False;
    nThreadStates = 6;
    strcpy(threadStates[0].stateLabel,"Dormant");
    threadStates[0].stateHandle = WindTaskDORMANT;
    strcpy(threadStates[1].stateLabel,"Delayed");
    threadStates[1].stateHandle = WindTaskDELAYED;
    strcpy(threadStates[2].stateLabel,"Pended");
    threadStates[2].stateHandle = WindTaskPENDED;
    strcpy(threadStates[3].stateLabel,"Suspended");
    threadStates[3].stateHandle = WindTaskSUSPENDED;
    strcpy(threadStates[4].stateLabel,"Ready");
    threadStates[4].stateHandle = WindTaskREADY;
    strcpy(threadStates[5].stateLabel,"Running");
    threadStates[5].stateHandle = WindTaskRUNNING;
}

// WindInterrupt -- A WIND extension of the VRTOS interrupt object
// used by the trampoline code to map the WIND semantics to
// Ck's.

extern "C" {

static int wind_isr_trampoline (ckvector_t vector,
				void *cookie)
{
    WindInterrupt *intr = (WindInterrupt *)cookie;
    intr->routine(intr->parameter);
    intr->setStatus(CK_INTR_ACKED);

    return CK_ISR_HANDLED;
}

}

WindInterrupt::WindInterrupt (ckvector_t _vector,
			      cklevel_t _level,
			      VOIDFUNCPTR _routine,
			      int _parameter,
			      const char *_name,
			      WindNode *_node) :
    CkInterrupt(_vector,
		_level,
		&wind_isr_trampoline,
		NULL,
		this,
		_name,
		_node)
{
    routine = _routine;
    parameter = _parameter;
}
