/*
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the WIND kernel synchronization objects
 * such as mutexes, semaphores, and message queues.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "CkManager.h"
#include "CkSynch.h"
#include "WindSynch.h"
#include "WindNode.h"
#include "WindTask.h"

WindSynch::WindSynch (const char *_name,
		      int _options,
		      InsertMode _mode,
		      CkStateExportObject *_synchState,
		      WindNode *_node) :
    CkSynchro(_name,
	      _mode,
	      _synchState,
	      _node),
    options(_options) {
}

void WindSynch::blockTask (WindTask *task, ckticks_t timeout)

{
    insert(task,task->getPri());
    getNode()->suspendThread(task,CKPEND,timeout,this);
}

void WindSynch::unblockTask (WindTask *task) {

    getNode()->readyThread(task,CKPEND);
}

void WindSynch::forget (CkThread *thread) {

    CkSynchro::forget(thread);
}

// WindSema4 -- Semaphore object supporting the binary and
// counting operating modes. Mutex semaphores are mapped on
// WindMutex objects due to the different semantics.

const char *WindSema4::stateArray[3] = {
    "IDLE", "PENDED", "POSTED"
};

WindSema4::WindSema4 (int _options,
		      int _count,
		      WindNode *_node) :
    WindSynch(NULL,
	      _options,
	      (_options & SEM_Q_PRIORITY) ? _node->getPriMode() : FIFO,
	      NULL,
	      _node)
{
    options = _options;
    count = _count;

    _node->nameObject(this,WIND_SEMA4_INODE);
    protoSetName(CString().format("%csem#%.3u",
				  (options & WIND_SEM_BINARY) ? 'b' : 'c',
				  getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "Semaphores",
					 stateArray,3);
    synchState->setState(_count > 0 ? WindSema4POSTED : WindSema4IDLE);
    protoInit();
}

WindSema4::~WindSema4 ()

{
    WindTask *task;
    
    while ((task = (WindTask *)get()) != NULL)
	{
	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void WindSema4::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("vxworks","WindSema4"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

const char *WindSema4::getTypedName ()

{
    CkSynchro::typedName.set(testOption(WIND_SEM_BINARY) ? "bsem" : "csem");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void WindSema4::dynamicExpose ()

{
    TclList __tkl, _tkl, tkl;

    _tkl.set("waiters");
    buildPendList(__tkl);
    _tkl.append(__tkl);
    tkl.append(_tkl);

    _tkl.set("value");
    _tkl.append(count);
    
    tkl.append(_tkl);
    
    protoInfo(SXPROTO_OBJECT_INFO,tkl.get(),tkl.length());
}

void WindSema4::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"value") == 0)
	    {
	    int newCount = atoi(parser.next());

	    if (newCount > count)
		{
		for (int relCount = newCount - count;
		     relCount > 0; relCount--)
		    release();
		}
	    else
		count = newCount;
	    }
	}
}

int WindSema4::acquire (WindTask *task, ckticks_t timeout)

{
    if (count > 0)
	{
	if (--count == 0)
	    synchState->setState(WindSema4IDLE);

	return CK_SYNCH_DONE;
	}

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    blockTask(task,timeout);
    synchState->setState(WindSema4PENDED);

    return CK_SYNCH_WAIT;
}

int WindSema4::release ()

{
    WindTask *task = (WindTask *)get();

    if (task)
	{
	unblockTask(task);

	if (getCount() == 0)
	    synchState->setState(WindSema4IDLE);

	return CK_SYNCH_RESCHED;
	}

    if (count + 1 != 0) // Check for roll-over.
	{
	count++;
	synchState->setState(WindSema4POSTED);
	}
    else
	CkManager::This->warning("semaphore count %s is rolling over!",
				 protoGetName());

    return CK_SYNCH_DONE;
}

int WindSema4::broadcast ()

{
    int s = getCount() > 0 ? CK_SYNCH_RESCHED : CK_SYNCH_DONE;
    WindTask *task;

    while ((task = (WindTask *)get()) != NULL)
	unblockTask(task);

    if (s == CK_SYNCH_RESCHED)
	synchState->setState(WindSema4IDLE);

    return s;
}

void WindSema4::forget (CkThread *thread)

{
    WindSynch::forget(thread);

    if (getCount() == 0)
	synchState->setState(WindSema4IDLE);
}

// WindMutex -- Mutex with optional support for Priority Inheritance
// Protocol.

const char *WindMutex::stateArray[3] = {
    "FREE", "LOCKED", "PENDED"
};

WindMutex::WindMutex (int _options,
		      WindNode *_node) :
    WindSynch(NULL,
	      _options,
	      (_options & SEM_Q_PRIORITY) ? _node->getPriMode() : FIFO,
	      NULL,
              _node)
{
    owner = NULL;
    ownCount = 0;

    _node->nameObject(this,WIND_MUTEX_INODE);
    protoSetName(CString().format("mutex#%.3u",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "Mutexes",
					 stateArray,3);
    synchState->setState(WindMutexFREE);
    protoInit();
}

WindMutex::~WindMutex ()

{
    WindTask *task;
    
    while ((task = (WindTask *)get()) != NULL)
	{
	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void WindMutex::protoInit ()

{
    // First, export mutex as a node-based object
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("vxworks","WindMutex"));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

void WindMutex::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("state");

    if (synchState->getState() == WindMutexFREE)
	_tclist.append("FREE");
    else
	_tclist.append("LOCKED");

    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void WindMutex::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data);
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"state") == 0)
	    {
	    const char *newState = parser.next();

	    if (!strcmp(newState,"FREE"))
		{
		if (owner)
		    unlock(owner);
		}
	    else
		synchState->setState(WindMutexLOCKED);
	    }
	}
}

const char *WindMutex::getTypedName ()

{
    CkSynchro::typedName.set("mutex");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

int WindMutex::lock (WindTask *task,
		     ckticks_t timeout)
{
    if (synchState->getState() == WindMutexFREE)
	{
	if (testOption(SEM_INVERSION_SAFE))
	    pipDeclare(task);

	owner = task;
	ownCount++;
	synchState->setState(WindMutexLOCKED);
	owner->enterSafeSection();

	return CK_SYNCH_DONE;
	}
    else if (task == owner)
	{
	ownCount++;
	return CK_SYNCH_DONE;	// Recursive lock.
	}

    pipApply(task);
    blockTask(task,timeout);
    synchState->setState(WindMutexPENDED);

    return CK_SYNCH_WAIT;
}

int WindMutex::unlock (WindTask *task)

{
    if (owner != task ||
	synchState->getState() == WindMutexFREE)
	return CK_SYNCH_ERROR;

    if (--ownCount > 0)
	return CK_SYNCH_DONE;

    // Lock the scheduling as exitSafeSection()
    // indirectly calls CkNode::resched().

    task->lock();

    owner->exitSafeSection();

    int s = pipExit();	// Returns CK_SYNCH_DONE if !inversion-safe

    owner = (WindTask *)get();

    if (owner)
	{
	ownCount++;

	unblockTask(owner);

	if (getCount() == 0)
	    synchState->setState(WindMutexLOCKED);

	owner->enterSafeSection();

	s = CK_SYNCH_RESCHED;
	}
    else
	synchState->setState(WindMutexFREE);

    task->unlock();

    return s;
}

void WindMutex::forget (CkThread *thread)

{
    WindSynch::forget(thread);

    if (getCount() == 0)
	synchState->setState(WindMutexLOCKED);
}

void WindMutex::safeSync (WindTask *waiter)

{
    if (testOption(SEM_DELETE_SAFE) && owner && waiter != owner)
	owner->safeSync(waiter);
}

// WIND message queue

const char *WindMsgQ::stateArray[4] = {
    "EMPTY", "PENDED", "POSTED", "FULL"
};

WindMsgQ::WindMsgQ (int _options,
		    unsigned _maxMsg,
		    unsigned _maxLength,
		    WindNode *_node) :
    WindSynch(NULL,
	      _options,
	      (_options & MSG_Q_PRIORITY) ? _node->getPriMode() : FIFO,
	      NULL,
	      _node), freeq(), inq()
{
    maxMsg = _maxMsg;
    maxLength = _maxLength;
    sendBlockCount = 0;

    // Preallocate the message space to speed up the operations
    // on the queue.

    for (unsigned n = 0; n < _maxMsg; n++)
	{
	WindMessage *msg = (WindMessage *)new char[sizeof(*msg) + _maxLength];
	freeq.append(msg);
	}

    _node->nameObject(this,WIND_MQ_INODE);
    protoSetName(CString().format("mq#%.3u",getCkId().getSeq()));
    protoSetExportable();
    synchState = new CkStateExportObject(_node,
					 protoGetName(),
					 "Message queues",
					 stateArray,4);
    synchState->setState(WindMsgQEMPTY);

    protoInit();
}

WindMsgQ::~WindMsgQ ()

{
    WindTask *task;
    
    while ((task = (WindTask *)get()) != NULL)
	{
	if (task->tstStatus(WIND_PEND_SENDSIDE))
	    {
	    dispose(task->msgGet());
	    task->msgSet(NULL);
	    }

	task->setStatus(CKRMID);
	unblockTask(task);
	}

    getNode()->unnameObject(this);
}

void WindMsgQ::protoInit ()

{
    CkObjectExportMessage em(getNode()->getNid(),
			     this,
			     getPlugInfo("vxworks",
					 "WindMsgQ",
					 protoGetName()));
    protoExport(&em,sizeof(em));

    synchState->protoInit();
}

const char *WindMsgQ::getTypedName ()

{
    CkSynchro::typedName.set("mq");
    CkSynchro::typedName.append(protoGetName());
    return CkSynchro::typedName;
}

void WindMsgQ::dynamicExpose ()

{
    TclList __tclist, _tclist, tclist;

    _tclist.set("waiters");
    buildPendList(__tclist);
    _tclist.append(__tclist);
    tclist.append(_tclist);

    _tclist.set("messages");
    __tclist.clear();

    for (WindMessage *msg = inq.first(); msg; msg = msg->link)
	{
	int nwords = Min(sizeof(u_long) * 4,msg->length / sizeof(u_long));
	TclList ___tclist;

	for (u_long *dstart = (u_long *)msg->data,
		 *dend = (u_long *)msg->data + nwords;
	     dstart < dend; dstart++)
	    ___tclist.appendx(*dstart);

	___tclist.append(CString().format("(sz=%u)",msg->length));
	___tclist.append(msg->time.format(NULL));
	__tclist.append(___tclist);
	}

    _tclist.append(__tclist);
    tclist.append(_tclist);
    
    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void WindMsgQ::dynamicConfigure (StProtoInfoMessage *mbuf) {
}

int WindMsgQ::send (WindTask *sender,
		    void *message,
		    unsigned length,
		    int options,
		    ckticks_t timeout)
{
    WindMessage *msg = freeq.pop();

    if (msg) // Can send without waiting?
	{
	msg->time = SxClock;
	msg->length = length;
	msg->flags = options;
	memcpy(msg->data,message,length);

	WindTaskIterator it(this);
	WindTask *task;

	while ((task = it.next()) != NULL)
	    {
	    // Search for a task pending on the receive side of this mq.
	    if (!task->tstStatus(WIND_PEND_SENDSIDE))
		{
		it.pop();
		break;
		}
	    }

	if (task)
	    {
	    // A task currently waits for a message -- pass it this
	    // message and wake it up.
	    task->msgSet(msg);
	    unblockTask(task);

	    if (getCount() == 0) // pended => empty?
		synchState->setState(WindMsgQEMPTY);

	    return CK_SYNCH_RESCHED;
	    }

	// Nobody waits for this message -- just enqueue it.

	if (options & MSG_PRI_URGENT)
	    inq.prepend(msg);
	else
	    inq.append(msg);

	if (inq.getCount() == 1) // empty => posted?
	    synchState->setState(WindMsgQPOSTED);

	return CK_SYNCH_DONE;
	}

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    if (sendBlockCount++ == 0)
	synchState->setState(WindMsgQFULL);

    // Save the message in a private buffer.
    msg = (WindMessage *)new char[sizeof(*msg) + maxLength];
    msg->time = SxClock;
    msg->length = length;
    msg->flags = options|WIND_MSG_PRIVATE;
    memcpy(msg->data,message,length);

    // 'sender' cannot be null at this point.

    sender->msgSet(msg);
    sender->setStatus(WIND_PEND_SENDSIDE);
    blockTask(sender,timeout);

    return CK_SYNCH_WAIT;
}

int WindMsgQ::receive (WindTask *receiver,
		       ckticks_t timeout)
{
    WindMessage *msg = inq.pop();

    if (msg)
	{
	receiver->msgSet(msg);

	if (sendBlockCount > 0)
	    {
	    WindTaskIterator it(this);
	    WindTask *task;

	    while ((task = (WindTask *)it.next()) != NULL)
		{
		// Search for a task waiting on the send-side.

		if (task->tstStatus(WIND_PEND_SENDSIDE))
		    {
		    it.pop();
		    break;
		    }
		}

	    // Transfer the message waiting on the send-side to
	    // the receive-side of the queue and unblock the sender.

	    msg = task->msgGet();

	    if (msg->flags & MSG_PRI_URGENT)
		inq.prepend(msg);
	    else
		inq.append(msg);

	    task->clrStatus(WIND_PEND_SENDSIDE);
	    unblockTask(task);

	    if (--sendBlockCount == 0)
		synchState->setState(WindMsgQPOSTED);

	    return CK_SYNCH_RESCHED;
	    }
	else if (inq.getCount() == 0) // posted => empty?
	    synchState->setState(WindMsgQEMPTY);

	return CK_SYNCH_DONE;
	}

    receiver->msgSet(NULL);

    if (timeout == CK_NONBLOCK)
	return CK_SYNCH_NOWAIT;

    blockTask(receiver,timeout);

    if (getCount() == 0) // empty => pended?
	synchState->setState(WindMsgQPENDED);

    return CK_SYNCH_WAIT;
}

void WindMsgQ::forget (CkThread *thread)

{
    CkSynchro::forget(thread);

    if (thread->tstStatus(WIND_PEND_SENDSIDE))
	{
	dispose(((WindTask *)thread)->msgGet());
	((WindTask *)thread)->msgSet(NULL);
	thread->clrStatus(WIND_PEND_SENDSIDE);
	sendBlockCount--;
	}

    if (getCount() == 0) // pended => empty?
	synchState->setState(WindMsgQEMPTY);
}
