/*
 * This file is part of the VxWorks simulation model for CarbonKernel.
 * Copyright (C) 2001  Philippe Gerum.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Description: Implementation of the WIND task.
 *
 * Author(s): rpm
 *
 * Contributor(s):
 *
 */

#ifdef __GNUG__
#pragma implementation
#endif // __GNUG__
#include <config.h>
#include <stdlib.h>
#include "CkManager.h"
#include "WindNode.h"
#include "WindTask.h"

WindTaskGList allWindTasks;

u_long WindTask::ptdKeySlots = 0;

u_long WindTask::seqGen = 0;

const char *WindTask::stateArray[6] = {

    "DORMANT",
    "DELAYED",
    "PENDED",
    "SUSPENDED",
    "READY",
    "RUNNING"
};

WindTask::WindTask (WIND_TCB *_pTcb,
		    const char *_name,
		    int _priority,
		    FUNCPTR _startAddr,
		    int _options,
		    int _startArgs[10],
		    WindNode *_node,
		    int _stackSize) :
    CkThread(_node,
	     _name,
	     _priority,
	     new CkThreadState(_name,
			       "Tasks",
			       WindTask::stateArray,6,
			       _node,
			       this),
	     _stackSize), // SIMEX default is 8k
    safeMutex(NULL,_node),
    safeCv(NULL,_node)
{
    startAddr = _startAddr;

    for (int n = 0; n < 10; n++)
	startArgs[n] = _startArgs[n];

    options = _options;
    startPri = _priority;
    pTcb = _pTcb;

    lastStateChangeTime = 0;
    safeLocks = 0;

    stateDistribs[0] = new CkHistogram(_name,"Delayed state",getNode());
    stateDistribs[0]->protoInit();
    stateDistribs[1] = new CkHistogram(_name,"Pended state",getNode());
    stateDistribs[1]->protoInit();
    stateDistribs[2] = new CkHistogram(_name,"Suspended state",getNode());
    stateDistribs[2]->protoInit();
    stateDistribs[3] = new CkHistogram(_name,"Ready state",getNode());
    stateDistribs[3]->protoInit();
    stateDistribs[4] = new CkHistogram(_name,"Running state",getNode());
    stateDistribs[4]->protoInit();

    _node->nameObject(this,WIND_TASK_INODE);

    // If no TCB was given (usual case, since taskSpawn() should be
    // preferred to taskInit() + taskActivate()), use the internal TCB.

    if (!pTcb)
	pTcb = &tcb;

    // Now, fill in the control block.

    pTcb->name = protoGetName();
    pTcb->options = _options;
    pTcb->status = 0;
    pTcb->priority = _priority;
    pTcb->entry = _startAddr;
    pTcb->errorStatus = OK;
    pTcb->handle = getCkId();	// i.e. Back pointer to the WindTask object.

    catchAsr(CK_INTERNAL_ASR,0,0);
    allWindTasks.append(this);
    protoInit();
}

WindTask::~WindTask ()

{
    allWindTasks.remove(this);
    getNode()->unnameObject(this);
}

void WindTask::body () {

    kroot(run)();
}

void WindTask::kroot(run) ()

{
    xtry(restartContext);

    getNode()->reinstateThread(this);

    setStatus(WIND_TASK_STARTED);

    startAddr(startArgs[0],
	      startArgs[1],
	      startArgs[2],
	      startArgs[3],
	      startArgs[4],
	      startArgs[5],
	      startArgs[6],
	      startArgs[7],
	      startArgs[8],
	      startArgs[9]);

    getNode()->kdoor(taskExit)(0);
}

void WindTask::startAsr (cksigmask_t sigs)

{
}

void WindTask::restart ()

{
    // Break the thread out of any wait it is currently in.
    getNode()->unblockThread(this);

    // If the task has been explicitely suspended, resume it.
    if (tstStatus(CKSUSP))
	getNode()->readyThread(this,CKSUSP);

    // Reset task priority to the initial one.
    setPri(startPri);

    // Clear pending signals.
    clrSignals();

    // Reset task's interrupt mask.
    setThreadIL(0);

    // Cancel pending alarms for this task.
    getNode()->cancelClientAlarms(this);

    // Reset other variables.
    safeLocks = 0;

    if (tstStatus(WIND_TASK_STARTED))
	{
	// Reset frame tracking variables at CkRunnable level.
	resetTracking();
	// Never returns if called on behalf of the running task.
	xraise(restartContext);
	}
}

void WindTask::logTaskState (WindTaskStates s)

{
    int olds = getThreadState();

    ITime dtLap = SxClock - lastStateChangeTime;

    if (olds >= 1 && olds <= 5)
	// <!> Depends on WindTaskStates enum order.
	stateDistribs[olds - 1]->add(dtLap);

    lastStateChangeTime = SxClock;

    setThreadState(s);
}

void WindTask::setReady () {

    logTaskState(WindTaskREADY);
}

void WindTask::setSuspended () {

    logTaskState(WindTaskSUSPENDED);
}

void WindTask::setPending () {

    logTaskState(WindTaskPENDED);
}

void WindTask::setRunning () {

    logTaskState(WindTaskRUNNING);
}

void WindTask::setDelayed () {

    logTaskState(WindTaskDELAYED);
}

void WindTask::setHalted () {

    setDormant();
}

void WindTask::setDormant () {

    logTaskState(WindTaskDORMANT);
}

void (*WindTask::getStartAddr ())(...) {

    return (void (*)(...))startAddr;
}

void WindTask::protoInit ()

{
    tstate->protoInit();
    CkThreadExportMessage tem(this,"vxworks","WindTask");
    protoExport(&tem,sizeof(tem));
}

void WindTask::dynamicExpose ()

{
    TclList _tclist, tclist;

    _tclist.set("prio");
    _tclist.append(getPri());
    tclist.append(_tclist);

    _tclist.set("ilevel");
    _tclist.append(getThreadIL());
    tclist.append(_tclist);

    _tclist.set("state");
    _tclist.append(stateArray[getThreadState()]);
    tclist.append(_tclist);

    if (tstStatus(CKPEND))
	{
	// update "wait channel" information using the
	// first pended synchronization object.
	_tclist.set("wchan");
	// If task is pending on a resource, add its name
	_tclist.append(getPendedSynchro()->first()->getTypedName());
	tclist.append(_tclist);
	}
    
    if (tstStatus(CKDELAY))
	{
	_tclist.set("timeout");
	_tclist.append(getTimerCount());
	tclist.append(_tclist);
	}

    protoInfo(SXPROTO_OBJECT_INFO,tclist.get(),tclist.length());
}

void WindTask::dynamicConfigure (StProtoInfoMessage *mbuf)

{
    TclListParser parser(mbuf->data); // protoInfo msg are null-terminated
    const char *setting;

    while ((setting = parser.next()) != NULL)
	{
	if (strcmp(setting,"prio") == 0)
	    {
	    int prio = atoi(parser.next());
	    getNode()->reniceThread(this,prio);
	    }
	else if (strcmp(setting,"ilevel") == 0)
	    {
	    int ilevel = atoi(parser.next());
	    setThreadIL(ilevel);
	    }
	}
}

// Per-task data management -- We don't mix with CkRunnable's
// key slots because we want to preserve the standard WIND namespace
// for per-task data indices and still be able to use our internal
// runnable-specific slots.

int WindTask::allocPtdKey ()

{
    if (ptdKeySlots == 0xffffffff)
	CkManager::This->fatal("no more per-task data index available");

    u_long mask = 1;
    int key = 0;

    while (ptdKeySlots & mask)
	key++, mask <<= 1;

    ptdKeySlots |= mask;

    return key;
}

void WindTask::freePtdKey (int key)

{
    u_long mask = (1 << key);
    ptdKeySlots &= ~mask;
}

void WindTask::enterSafeSection ()

{
    if (safeLocks++ > 0)
	{
	if (safeLocks & 0xff000000) // FIXME: Is there a maximum lock count?
	    CkManager::This->warning("taskSafe(): deletion overlocked");

	return;
	}

    // Enforcing the critical section is useless on behalf on
    // non-instrumented code since it cannot be preempted.
    // This is only done to make the model trigger the expected
    // synchronization events.
    safeMutex.lock(this,0);
    setStatus(WIND_TASK_SAFE);
    safeMutex.unlock(this);
}

void WindTask::exitSafeSection ()

{
    // FIXME: It seems that VxWorks silently checks for
    // spurious safe section exits.

    if (!safeLocks || --safeLocks > 0)
	return;

    // Same comment as above.
    safeMutex.lock(this,0);
    clrStatus(WIND_TASK_SAFE);
    safeCv.broadcast();	// Unblock the waiters.
    safeMutex.unlock(this);
}

void WindTask::safeSync (WindTask *waiter)

{
    safeMutex.lock(waiter,0);

    while (tstStatus(WIND_TASK_SAFE))
	safeCv.wait(waiter,&safeMutex,CK_INFINITE);

    safeMutex.unlock(waiter);
}
